from django.test import TestCase


class BaseTestCase(TestCase):
    def assertObjectSerializesTo(self, obj, subset):
        serialized = obj.serialize()
        self.assertDictContainsSubset(subset, serialized)


from django.contrib.auth import get_user_model
from django.test import RequestFactory
from wagtail.contrib.modeladmin.options import ModelAdmin
from django_webtest import WebTest
from shop.models import Product
from admin_list_controls.views import ListControlsIndexView


User = get_user_model()


class TestViews(WebTest):
    def setUp(self):
        self.factory = RequestFactory()
        self.superuser = User.objects.create_superuser(
            username='test',
            email='test@example.com',
            password='test',
        )

    def tearDown(self):
        User.objects.all().delete()

    def test_view_init_and_context(self):
        class TestView(ListControlsIndexView):
            pass

        view = self.list_view_class_to_view_function(TestView)
        response = view(self.create_superuser_request('/'))
        self.assertIn('admin_list_controls', response.context_data)
        self.assertIsInstance(response.context_data['admin_list_controls'], dict)

    def create_superuser_request(self, url):
        request = self.factory.get(url)
        request.user = self.superuser
        return request

    def instantiate_list_view_class(self, list_view_class) -> ListControlsIndexView:
        class TestModelAdmin(ModelAdmin):
            model = Product
            index_view_class = list_view_class

        return list_view_class(model_admin=TestModelAdmin())

    def list_view_class_to_view_function(self, list_view_class):
        class TestModelAdmin(ModelAdmin):
            model = Product
            index_view_class = list_view_class

        request = self.factory.get('/')
        request.user = self.superuser
        return TestModelAdmin().index_view



from admin_list_controls.components import ListControls, Block, Spacer, Columns, Divider, Panel, \
    Icon, Text, Button, Summary
from admin_list_controls.tests.utils import BaseTestCase


class TestComponents(BaseTestCase):
    def test_list_controls_component(self):
        self.assertObjectSerializesTo(
            ListControls(),
            {
                'object_type': 'list_controls',
                'children': None,
                'style': None,
            },
        )
        serialized = ListControls().serialize()
        self.assertEqual(serialized['extra_classes'], None)
        self.assertIn('ListControls', serialized['component_id'])
        nested = ListControls()
        self.assertObjectSerializesTo(
            ListControls()(nested),
            {
                'object_type': 'list_controls',
                'children': [nested.serialize()],
            },
        )

    def test_block_component(self):
        self.assertObjectSerializesTo(
            Block(),
            {
                'object_type': 'block',
            },
        )

    def test_spacer_component(self):
        self.assertObjectSerializesTo(
            Spacer(),
            {
                'object_type': 'block',
                'style': {'paddingTop': '20px'},
            },
        )

    def test_columns_component(self):
        component = Columns()(
            [Text('Column 1')],
            ['Column 2'],
        )
        component.prepare_children()
        serialized = component.serialize()
        self.assertEqual(serialized['object_type'], 'columns')
        self.assertEqual(len(serialized['children']), 2)
        self.assertDictContainsSubset(
            {
                'object_type': 'block',
                'style': {
                    'width': 'calc((100% - ((2 - 1) * 10px)) / 2)',
                    'float': 'left',
                    'marginLeft': '0',
                },
            },
            serialized['children'][0],
        )
        self.assertDictContainsSubset(
            {
                'object_type': 'text',
                'content': 'Column 1',
            },
            serialized['children'][0]['children'][0],
        )
        self.assertDictContainsSubset(
            {
                'object_type': 'block',
                'style': {
                    'width': 'calc((100% - ((2 - 1) * 10px)) / 2)',
                    'float': 'left',
                    'marginLeft': '10px',
                },
            },
            serialized['children'][1],
        )
        self.assertDictContainsSubset(
            {
                'object_type': 'text',
                'content': 'Column 2',
            },
            serialized['children'][1]['children'][0],
        )

    def test_divider_component(self):
        self.assertObjectSerializesTo(
            Divider(),
            {
                'object_type': 'divider',
            },
        )

    def test_panel_component(self):
        self.assertObjectSerializesTo(
            Panel(),
            {
                'object_type': 'panel',
                'collapsed': False,
            },
        )
        self.assertObjectSerializesTo(
            Panel(collapsed=True),
            {
                'object_type': 'panel',
                'collapsed': True,
            },
        )
        self.assertObjectSerializesTo(
            Panel(collapsed=False),
            {
                'object_type': 'panel',
                'collapsed': False,
            },
        )

    def test_icon_component(self):
        self.assertObjectSerializesTo(
            Icon(classes='test_class_name'),
            {
                'object_type': 'icon',
                'extra_classes': 'test_class_name',
            },
        )

    def test_text_component(self):
        self.assertObjectSerializesTo(
            Text(''),
            {
                'object_type': 'text',
                'content': '',
                'size': 'medium',
            },
        )
        self.assertObjectSerializesTo(
            Text('test'),
            {
                'object_type': 'text',
                'content': 'test',
                'size': 'medium',
            },
        )
        self.assertObjectSerializesTo(
            Text('test', size=Text.MEDIUM),
            {
                'object_type': 'text',
                'content': 'test',
                'size': 'medium',
            },
        )
        self.assertObjectSerializesTo(
            Text('test', size=Text.LARGE),
            {
                'object_type': 'text',
                'content': 'test',
                'size': 'large',
            },
        )

    def test_button_component(self):
        self.assertObjectSerializesTo(
            Button(),
            {
                'object_type': 'button',
                'action': [],
            },
        )

        class FakeAction:
            def serialize(self):
                return 'serialized_fake_action'

        self.assertObjectSerializesTo(
            Button(action=FakeAction()),
            {
                'object_type': 'button',
                'action': ['serialized_fake_action'],
            },
        )

        self.assertObjectSerializesTo(
            Button(action=[FakeAction(), FakeAction()]),
            {
                'object_type': 'button',
                'action': ['serialized_fake_action', 'serialized_fake_action'],
            },
        )

    def test_summary_component(self):
        self.assertObjectSerializesTo(
            Summary(),
            {
                'object_type': 'summary',
                'summary': [],
                'reset_label': 'Reset all',
            },
        )

    def test_string_children_are_converted_to_text_components(self):
        serialized = ListControls()(
            'test string 1',
            'test string 2'
        ).serialize()
        self.assertDictContainsSubset(
            {
                'object_type': 'text',
                'content': 'test string 1',
            },
            serialized['children'][0],
        )
        self.assertDictContainsSubset(
            {
                'object_type': 'text',
                'content': 'test string 2',
            },
            serialized['children'][1],
        )



import os
import json
from collections import Iterable

from django.conf import settings
from django.utils.safestring import mark_safe
from wagtail.contrib.modeladmin.views import IndexView
from .components import ListControls
from .selectors import LayoutSelector
from .vendor import webpack_manifest


class ListControlsIndexViewMixin:
    """
    Adds filtering, sorting and other controls to a modeladmin's list/index view
    """

    _built_list_controls = None
    _items_per_page = None
    _has_prepared_list_controls = False

    def build_list_controls(self):
        """
        A hook provided to declare the filters for the UI
        """
        return ListControls()

    def apply_list_controls_to_queryset(self, queryset):
        """
        A hook provided to modify the queryset
        """
        for obj in self.get_list_controls().flatten_tree():
            if hasattr(obj, 'apply_to_queryset'):
                queryset = obj.apply_to_queryset(queryset)
        return queryset

    def get_list_controls(self):
        if not self._built_list_controls:
            list_controls = self.build_list_controls()
            # Allow the controls to be defined in a list
            if isinstance(list_controls, Iterable):
                list_controls = ListControls()(*list_controls)
            self._built_list_controls = list_controls
        return self._built_list_controls

    def get_selected_list_control_layout(self):
        for obj in self.get_list_controls().flatten_tree():
            if isinstance(obj, LayoutSelector) and obj.is_selected:
                return obj

    def get_template_names(self):
        return ['admin_list_controls/index.html'] + super().get_template_names()

    def get_filters_params(self, params=None):
        """
        Prevent the modeladmin's built-in querying from raising errors when it encounters
        our custom GET data
        """
        params = super().get_filters_params(params)

        for obj in self.get_list_controls().flatten_tree():
            name = getattr(obj, 'name', None)
            if name and name in params:
                del params[name]
        return params

    def get_search_results(self, *args, **kwargs):
        queryset = super().get_search_results(*args, **kwargs)

        self.prepare_list_controls()
        return self.apply_list_controls_to_queryset(queryset)

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)

        selected_layout = self.get_selected_list_control_layout()
        if selected_layout and selected_layout.template:
            selected_layout_template = selected_layout.template
        else:
            selected_layout_template = None

        context_data['admin_list_controls'] = {
            'index_template': self.get_list_controls_index_template(),
            # Consumed by the front-end code to build the UI
            'initial_state': json.dumps({
                'admin_list_controls': self.get_list_controls().serialize(),
            }),
            'selected_layout_template': selected_layout_template,
            'widget_js': self.get_list_controls_widget_js(),
        }

        return context_data

    def get_list_controls_index_template(self):
        return 'modeladmin/index.html'

    def get_list_controls_widget_js(self):
        manifest_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), 'webpack_manifest.json',
        )
        manifest = webpack_manifest.load(
            path=manifest_path,
            static_url=settings.STATIC_URL,
            static_root=settings.STATIC_ROOT,
            debug=settings.DEBUG,
        )
        return mark_safe(str(manifest.admin_list_controls.js))

    def prepare_list_controls(self):
        # The lifecycle hooks we use to wire into the request/response are sometimes
        # executed multiple times per-request. To avoid odd bugs (components wrapping
        # other components multiple times, etc), we need to manually enforce a single
        # execution of the following method
        if self._has_prepared_list_controls:
            return
        self._has_prepared_list_controls = True

        list_controls = self.get_list_controls()

        controls_by_name = {}
        for obj in list_controls.flatten_tree():
            # Allow objects
            if hasattr(obj, 'handle_request'):
                obj.handle_request(self.request)
            name = getattr(obj, 'name', None)
            if name:
                if name not in controls_by_name:
                    controls_by_name[name] = []
                controls_by_name[name].append(obj)

        # If no selectors have been selected for a particular param,
        # indicate the default selectors should be used
        for name, controls in controls_by_name.items():
            selector_controls = [
                control for control in controls
                if control.object_type == 'selector'
            ]
            has_selected = False
            for control in selector_controls:
                if control.object_type == 'selector' and control.is_selected:
                    has_selected = True
                    break
            if not has_selected:
                for control in selector_controls:
                    if control.is_default:
                        control.is_selected = True
                        break

        # Some objects wrap their children in other components
        for obj in list_controls.flatten_tree():
            if hasattr(obj, 'prepare_children'):
                obj.prepare_children()

        # Some objects are derived from the final state of the tree

        flattened_tree = list_controls.flatten_tree()
        for obj in flattened_tree:
            if hasattr(obj, 'derive_from_components'):
                obj.derive_from_components(flattened_tree)


class ListControlsIndexView(ListControlsIndexViewMixin, IndexView):
    pass



class BaseAction:
    object_type = 'action'
    action_type = ''

    def serialize(self):
        return {
            'object_type': self.object_type,
            'action_type': self.action_type,
        }


class SetValue(BaseAction):
    action_type = 'set_value'

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def serialize(self):
        return dict(super().serialize(), **{
            'name': self.name,
            'value': self.value,
        })


class RemoveValue(BaseAction):
    action_type = 'remove_value'

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def serialize(self):
        return dict(super().serialize(), **{
            'name': self.name,
            'value': self.value,
        })


class Link(BaseAction):
    action_type = 'link'

    def __init__(self, url):
        self.url = url

    def serialize(self):
        return dict(super().serialize(), **{
            'url': self.url,
        })


class TogglePanel(BaseAction):
    action_type = 'toggle_panel'
    show_panel_toggle_icon = True

    def __init__(self, ref):
        self.ref = ref

    def serialize(self):
        return dict(super().serialize(), **{
            'ref': self.ref,
            'show_panel_toggle_icon': self.show_panel_toggle_icon,
        })


class CollapsePanel(BaseAction):
    action_type = 'collapse_panel'

    def __init__(self, ref):
        self.ref = ref

    def serialize(self):
        return dict(super().serialize(), **{
            'ref': self.ref,
        })


class ClearSearchInput(BaseAction):
    action_type = 'clear_search_input'


class SubmitForm(BaseAction):
    action_type = 'submit_form'



from .components import BaseComponent
from .actions import RemoveValue, SubmitForm


class BaseFilter(BaseComponent):
    object_type = 'filter'
    filter_type = ''
    name = None
    cleaned_value = None
    can_have_children = False

    def __init__(
        self,
        name,
        label=None,
        apply_to_queryset=None,
        default_value=None,
        summary_label=None,
        exclude_default_value_from_summary=True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.name = name
        self.label = label
        self._apply_to_queryset = apply_to_queryset
        self.default_value = default_value
        self.summary_label = summary_label
        self.exclude_default_value_from_summary = exclude_default_value_from_summary

    def handle_request(self, request):
        self.cleaned_value = self.clean(request)
        if self.cleaned_value is None:
            self.cleaned_value = self.default_value

    def clean(self, request):
        return request.GET.get(self.name)

    def apply_to_queryset(self, queryset):
        if self._apply_to_queryset and self.cleaned_value:
            return self._apply_to_queryset(queryset, self.cleaned_value)
        return queryset

    def serialize_summary(self):
        if self.exclude_default_value_from_summary and self.cleaned_value == self.default_value:
            return
        return self.serialize_summary_for_value(self.cleaned_value)

    def serialize_summary_for_value(self, value):
        if value:
            label = self.summary_label
            if not label:
                label = self.label
            display_value = self.get_summary_display_value_for_value(value)

            return {
                'name': self.name,
                'label': label,
                'display_value': display_value,
                'value': value,
                'action': [
                    RemoveValue(name=self.name, value=value).serialize(),
                    SubmitForm().serialize(),
                ],
            }

    def get_summary_display_value_for_value(self, value):
        return value

    def serialize(self):
        return dict(super().serialize(), **{
            'filter_type': self.filter_type,
            'name': self.name,
            'label': self.label,
            'value': self.cleaned_value,
        })


class TextFilter(BaseFilter):
    filter_type = 'text'

    def clean(self, *args, **kwargs):
        value = super().clean(*args, **kwargs)
        if value is None:
            return ''
        return value


class BooleanFilter(BaseFilter):
    filter_type = 'boolean'

    def clean(self, *args, **kwargs):
        value = super().clean(*args, **kwargs)
        return bool(value)


class BaseChoiceFilter(BaseFilter):
    def __init__(self, choices, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.choices = choices

    def clean(self, request):
        whitelisted_values = [choice[0] for choice in self.choices]
        value = request.GET.get(self.name)
        if value in whitelisted_values:
            return value
        elif self.default_value:
            return self.default_value

    def serialize(self):
        return dict(super().serialize(), **{
            'choices': self.choices,
        })

    def get_summary_display_value_for_value(self, value):
        """
        Returns the corresponding display value for the raw value
        """
        choice_dict = dict(self.choices)
        return choice_dict.get(value, value)


class RadioFilter(BaseChoiceFilter):
    filter_type = 'radio'


class ChoiceFilter(BaseChoiceFilter):
    filter_type = 'choice'

    def __init__(self, choices, multiple=False, *args, **kwargs):
        super().__init__(choices=choices, *args, **kwargs)

        self.multiple = multiple

    def clean(self, request):
        if self.multiple:
            whitelisted_values = [choice[0] for choice in self.choices]
            values = request.GET.getlist(self.name)
            cleaned_values = [
                value for value in values
                if value in whitelisted_values
            ]
            if cleaned_values:
                return cleaned_values
            elif self.default_value:
                return [self.default_value]
            else:
                return []
        return super().clean(request)

    def serialize_summary(self):
        if self.multiple:
            summaries = []
            for value in self.cleaned_value:
                summaries.append(
                    self.serialize_summary_for_value(value))
            return summaries
        return super().serialize_summary()

    def serialize(self):
        return dict(super().serialize(), **{
            'multiple': self.multiple,
        })



from collections.abc import Iterable
from django.utils.translation import gettext as _

from .actions import SubmitForm, ClearSearchInput
from .exceptions import ConfigurationError


class BaseComponent:
    object_type = ''
    children = None
    can_have_children = True
    _children_defined = False

    def __init__(self, style=None, extra_classes=''):
        self.style = self.get_default_style()
        if style:
            self.style.update(style)
        self.extra_classes = extra_classes
        self.component_id = '%s-%s' % (type(self).__name__, id(self))

    def __call__(self, *args):
        return self.set_children(*args)

    def set_children(self, *args):
        if self._children_defined:
            raise ConfigurationError('Children have already been defined')
        if not self.can_have_children:
            raise ConfigurationError('%s cannot have children' % type(self))
        self._children_defined = True
        self.children = []
        for child in args:
            if isinstance(child, str):
                self.children.append(
                    Text(content=child))
            else:
                self.children.append(child)
        return self

    def serialize(self):
        serialized_children = None
        if self.children:
            serialized_children = [
                child.serialize()
                for child in self.children
            ]

        return {
            'component_id': self.component_id,
            'object_type': self.object_type,
            'children': serialized_children,
            'style': self.serialize_style(),
            'extra_classes': self.extra_classes,
        }

    def serialize_style(self):
        """
        Convert conventional css-props into the camel-case that React prefers
        """
        if not self.style:
            return None

        serialized = {}
        for key, value in self.style.items():
            if '-' in key:
                parts = key.split('-')
                converted = parts[0] + ''.join(part.title() for part in parts[1:])
                serialized[converted] = value
            else:
                serialized[key] = value
        return serialized

    def get_default_style(self):
        return {}

    def flatten_tree(self):
        components = [self]
        if self.children:
            for child in self.children:
                if isinstance(child, Iterable):
                    for nested_child in child:
                        components += nested_child.flatten_tree()
                else:
                    components += child.flatten_tree()
        return components


class ListControls(BaseComponent):
    object_type = 'list_controls'


class Block(BaseComponent):
    object_type = 'block'


class Spacer(Block):
    def get_default_style(self):
        return {'padding-top': '20px'}


class Columns(BaseComponent):
    object_type = 'columns'

    def __init__(self, column_count=2, **kwargs):
        super().__init__(**kwargs)
        self.column_count = column_count

    def prepare_children(self):
        self._original_children = self.children

        column_count = self.column_count
        gutter_width = '20px'
        column_width = 'calc((100% - (({cc} - 1) * {gw})) / {cc})'.format(
            cc=column_count,
            gw=gutter_width,

        )

        columns = [[] for _ in range(column_count)]
        next_column_index = 0
        for i, child in enumerate(self.children):
            next_column = columns[next_column_index]
            next_column.append(child)
            next_column_index += 1
            if (next_column_index + 1) > column_count:
                next_column_index = 0

        self.children = []
        for column_index, column_children in enumerate(columns):
            style = {
                'width': column_width,
                'float': 'left',
                'margin-left': gutter_width if column_index > 0 else '0',
            }
            self.children.append(
                Block(style)(*column_children),
            )


class Divider(BaseComponent):
    object_type = 'divider'
    can_have_children = False


class Panel(BaseComponent):
    object_type = 'panel'

    def __init__(self, ref=None, collapsed=False, **kwargs):
        super().__init__(**kwargs)
        self.ref = ref
        self.collapsed = collapsed

    def serialize(self):
        return dict(super().serialize(), **{
            'ref': self.ref,
            'collapsed': self.collapsed,
        })


class Icon(BaseComponent):
    object_type = 'icon'
    can_have_children = False

    def __init__(self, classes=None, **kwargs):
        if classes:
            kwargs['extra_classes'] = classes
        super().__init__(**kwargs)


class Text(BaseComponent):
    object_type = 'text'
    can_have_children = False

    # Sizes
    LARGE = 'large'
    MEDIUM = 'medium'

    def __init__(self, content, size=MEDIUM, **kwargs):
        super().__init__(**kwargs)
        self.content = content
        self.size = size

    def serialize(self):
        return dict(super().serialize(), **{
            'content': self.content,
            'size': self.size,
        })


class Button(BaseComponent):
    object_type = 'button'

    def __init__(self, action=None, **kwargs):
        super().__init__(**kwargs)
        if action is None:
            action = []
        elif not isinstance(action, Iterable):
            # Allow single actions to be defined without wrapping them in a list
            action = [action]
        self.action = action

    def serialize(self):
        serialized_action = []
        for action in self.action:
            serialized_action.append(action.serialize())
        return dict(super().serialize(), **{
            'action': serialized_action,
        })


class Summary(BaseComponent):
    object_type = 'summary'

    def __init__(
        self,
        reset_label=_('Reset all'),
        include_search_query=True,
        search_query_name='q',
        search_query_label=_('Search'),
        search_query_value=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.reset_label = reset_label
        self.summary = []
        self.include_search_query = include_search_query
        self.search_query_name = search_query_name
        self.search_query_label = search_query_label
        self.search_query_value = search_query_value

    def handle_request(self, request):
        if self.include_search_query:
            self.search_query_value = request.GET.get(self.search_query_name)

    def derive_from_components(self, components):
        # Include the content of the built-in search field
        if self.include_search_query and self.search_query_value:
            self.summary.append({
                'name': self.search_query_name,
                'label': self.search_query_label,
                'value': self.search_query_value,
                'action': [
                    ClearSearchInput().serialize(),
                    SubmitForm().serialize(),
                ]
            })

        for component in components:
            if hasattr(component, 'serialize_summary'):
                summaries = component.serialize_summary()
                if isinstance(summaries, dict):
                    summaries = [summaries]
                if summaries:
                    self.summary += summaries

    def serialize(self):
        return dict(super().serialize(), **{
            'summary': self.summary,
            'reset_label': self.reset_label,
        })


